class class_prop(object):
    """
    This decorator provide an easy to use interface to initialize and cache fixtures for test cases.
    With a little additional code for TestCase it can cache results of database operations for
    a whole test case.
    >>>@classmethod
    ...def setUpTestData(cls):
    ...    for name in dir(cls):
    ...        getattr(cls, name)

    Possible examples of usage:
    >>>class ExampleTestCase(TestCase):
    ...    @class_prop
    ...    def c(cls):
    ...        return cls.create_user('test_user')
    ...
    ...    c = class_prop(lambda cls: cls.create_user('user1'), name='c')
    Pay attention that decorated method behaves like a classmethod. Also if `class_prop` is used to
     decorate lambda function, it is necessary to provide additional keyword argument `name` with
     desired name for cached property.

    Also it can be used in test case mixins to provide collections of fixtures.
    >>>class TestUserMixing(object):
    ...    c1 = class_prop(lambda cls: cls.create_user('user1'), name='c1')
    ...    c2 = class_prop(lambda cls: cls.create_user('user2'), name='c2')
    """
    def __init__(self, f, name=None):
        self.f = f
        self.name = name

    def __get__(self, obj, klass=None):
        assert klass is not None, "Must not be used as an object attribute"

        result = self.f(klass)
        setattr(klass, self.name or self.f.__name__, result)
        return result



import logging
import sys
from py.path import local

import boto3

from fan_tools.backup import BackupMonitoring
from fan_tools.backup.utils import run_main


class S3Backup(BackupMonitoring):
    def __init__(self, bucket_name=None, **kwargs):
        # from google.cloud import storage
        super().__init__(**kwargs)
        assert bucket_name, 'You should define bucket-name'
        self.cli = boto3.resource('s3')
        self.bucket = self.cli.Bucket(bucket_name)

    def get_backups_list(self):
        items = list(self.bucket.objects.filter(Prefix=self.prefix))
        return [local(s.key).basename for s in items]

    def perform_upload(self, src_name, dst_name):
        self.log.debug('Upload {src_name} => {dst_name}')
        self.bucket.upload_file(src_name, dst_name)


def main():
    """
    Example setup for s3.

    BACKUP_COMMAND='touch /dumps/aaa; echo aaa' ENABLE_BACKUP=true fan_s3_backup
    """
    import os
    from fan_tools.fan_logging import setup_logger

    setup_logger('backup_monitoring', enable_stdout=True)

    os.environ['AWS_ACCESS_KEY_ID'] = os.environ['AWS_BACKUP_KEY']
    os.environ['AWS_SECRET_ACCESS_KEY'] = os.environ['AWS_BACKUP_SECRET']

    run_main(S3Backup, environ_defaults={'bucket': 'AWS_BACKUP_BUCKET'})


if __name__ == '__main__':
    main()



import logging
from functools import cached_property

from django.conf import settings
from opentelemetry.instrumentation.django.middleware import _DjangoMiddleware
from opentelemetry.trace import (
    INVALID_SPAN,
    INVALID_SPAN_CONTEXT,
    get_current_span,
    get_tracer,
    get_tracer_provider,
)

from fan_tools.fan_logging import JSFormatter


class OtelJSFormatter(JSFormatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._required_fields.extend(['otelSpanID', 'otelTraceID', 'otelServiceName'])


def instrument_logging(**kwargs):
    old_factory = logging.getLogRecordFactory()
    service_name = ""
    provider = kwargs.get("tracer_provider", None) or get_tracer_provider()
    resource = provider.resource if provider and hasattr(provider, 'resource') else None
    if resource:
        service_name = resource.attributes.get("service.name")

    def record_factory(*args, **kwargs):
        record = old_factory(*args, **kwargs)
        record.otelSpanID = "0"
        record.otelTraceID = "0"
        record.otelServiceName = service_name

        span = get_current_span()
        if span != INVALID_SPAN:
            ctx = span.get_span_context()
            if ctx != INVALID_SPAN_CONTEXT:
                record.otelSpanID = format(ctx.span_id, "016x")
                record.otelTraceID = format(ctx.trace_id, "032x")

                # datadog requires conversion
                record.__dict__['dd.span_id'] = str(ctx.span_id)
                record.__dict__['dd.trace_id'] = str(ctx.trace_id & 0xFFFFFFFFFFFFFFFF)

                # newrelic requires custom names
                record.__dict__['span.id'] = record.otelSpanID
                record.__dict__['trace.id'] = record.otelTraceID
        return record

    logging.setLogRecordFactory(record_factory)


class OtelDjangoMiddleware(_DjangoMiddleware):
    @cached_property
    def _tracer(self):
        tracer_provider = get_tracer_provider()
        tracer = get_tracer(
            'django',
            '0.1',
            tracer_provider=tracer_provider,
        )
        return tracer


def instrument_django():
    settings.MIDDLEWARE.insert(0, 'fan_tools.otel.OtelDjangoMiddleware')


def instrument_psycopg2():
    from opentelemetry.instrumentation.psycopg2 import Psycopg2Instrumentor

    Psycopg2Instrumentor().instrument()



import logging
import socket
from logging import Handler
from logging.handlers import SYSLOG_UDP_PORT


"""
Fixed version of logging.handers.SysLogHandler that supports reconnects
Can be removed after https://github.com/python/cpython/pull/26490 get merged

Usage:

    setup_logger(
        'cms',
        root_dir=log_dir,
        enable_stdout=True,
        stdout_level='INFO',
        json_formatter=formatter_name,
        handlers={'vector': {
            'level': 'DEBUG',
            'class': 'fan_tools.fan_logging.handlers.SysLogHandler',
            'address': ('vector', 9998),
            'formatter': 'json',
        }}
    )

"""

class SysLogHandler(logging.Handler):
    """
    A handler class which sends formatted logging records to a syslog
    server. Based on Sam Rushing's syslog module:
    http://www.nightmare.com/squirl/python-ext/misc/syslog.py
    Contributed by Nicolas Untz (after which minor refactoring changes
    have been made).
    """

    # from <linux/sys/syslog.h>:
    # ======================================================================
    # priorities/facilities are encoded into a single 32-bit quantity, where
    # the bottom 3 bits are the priority (0-7) and the top 28 bits are the
    # facility (0-big number). Both the priorities and the facilities map
    # roughly one-to-one to strings in the syslogd(8) source code.  This
    # mapping is included in this file.
    #
    # priorities (these are ordered)

    LOG_EMERG     = 0       #  system is unusable
    LOG_ALERT     = 1       #  action must be taken immediately
    LOG_CRIT      = 2       #  critical conditions
    LOG_ERR       = 3       #  error conditions
    LOG_WARNING   = 4       #  warning conditions
    LOG_NOTICE    = 5       #  normal but significant condition
    LOG_INFO      = 6       #  informational
    LOG_DEBUG     = 7       #  debug-level messages

    #  facility codes
    LOG_KERN      = 0       #  kernel messages
    LOG_USER      = 1       #  random user-level messages
    LOG_MAIL      = 2       #  mail system
    LOG_DAEMON    = 3       #  system daemons
    LOG_AUTH      = 4       #  security/authorization messages
    LOG_SYSLOG    = 5       #  messages generated internally by syslogd
    LOG_LPR       = 6       #  line printer subsystem
    LOG_NEWS      = 7       #  network news subsystem
    LOG_UUCP      = 8       #  UUCP subsystem
    LOG_CRON      = 9       #  clock daemon
    LOG_AUTHPRIV  = 10      #  security/authorization messages (private)
    LOG_FTP       = 11      #  FTP daemon
    LOG_NTP       = 12      #  NTP subsystem
    LOG_SECURITY  = 13      #  Log audit
    LOG_CONSOLE   = 14      #  Log alert
    LOG_SOLCRON   = 15      #  Scheduling daemon (Solaris)

    #  other codes through 15 reserved for system use
    LOG_LOCAL0    = 16      #  reserved for local use
    LOG_LOCAL1    = 17      #  reserved for local use
    LOG_LOCAL2    = 18      #  reserved for local use
    LOG_LOCAL3    = 19      #  reserved for local use
    LOG_LOCAL4    = 20      #  reserved for local use
    LOG_LOCAL5    = 21      #  reserved for local use
    LOG_LOCAL6    = 22      #  reserved for local use
    LOG_LOCAL7    = 23      #  reserved for local use

    priority_names = {
        "alert":    LOG_ALERT,
        "crit":     LOG_CRIT,
        "critical": LOG_CRIT,
        "debug":    LOG_DEBUG,
        "emerg":    LOG_EMERG,
        "err":      LOG_ERR,
        "error":    LOG_ERR,        #  DEPRECATED
        "info":     LOG_INFO,
        "notice":   LOG_NOTICE,
        "panic":    LOG_EMERG,      #  DEPRECATED
        "warn":     LOG_WARNING,    #  DEPRECATED
        "warning":  LOG_WARNING,
        }

    facility_names = {
        "auth":         LOG_AUTH,
        "authpriv":     LOG_AUTHPRIV,
        "console":      LOG_CONSOLE,
        "cron":         LOG_CRON,
        "daemon":       LOG_DAEMON,
        "ftp":          LOG_FTP,
        "kern":         LOG_KERN,
        "lpr":          LOG_LPR,
        "mail":         LOG_MAIL,
        "news":         LOG_NEWS,
        "ntp":          LOG_NTP,
        "security":     LOG_SECURITY,
        "solaris-cron": LOG_SOLCRON,
        "syslog":       LOG_SYSLOG,
        "user":         LOG_USER,
        "uucp":         LOG_UUCP,
        "local0":       LOG_LOCAL0,
        "local1":       LOG_LOCAL1,
        "local2":       LOG_LOCAL2,
        "local3":       LOG_LOCAL3,
        "local4":       LOG_LOCAL4,
        "local5":       LOG_LOCAL5,
        "local6":       LOG_LOCAL6,
        "local7":       LOG_LOCAL7,
        }

    #The map below appears to be trivially lowercasing the key. However,
    #there's more to it than meets the eye - in some locales, lowercasing
    #gives unexpected results. See SF #1524081: in the Turkish locale,
    #"INFO".lower() != "info"
    priority_map = {
        "DEBUG" : "debug",
        "INFO" : "info",
        "WARNING" : "warning",
        "ERROR" : "error",
        "CRITICAL" : "critical"
    }

    def __init__(self, address=('localhost', SYSLOG_UDP_PORT),
                 facility=LOG_USER, socktype=None):
        """
        Initialize a handler.

        If address is specified as a string, a UNIX socket is used. To log to a
        local syslogd, "SysLogHandler(address="/dev/log")" can be used.
        If facility is not specified, LOG_USER is used. If socktype is
        specified as socket.SOCK_DGRAM or socket.SOCK_STREAM, that specific
        socket type will be used. For Unix sockets, you can also specify a
        socktype of None, in which case socket.SOCK_DGRAM will be used, falling
        back to socket.SOCK_STREAM.
        """
        logging.Handler.__init__(self)

        self.address = address
        self.facility = facility
        self.socktype = socktype
        self.socket = None
        self.createSocket()

    def _connect_unixsocket(self, address):
        use_socktype = self.socktype
        if use_socktype is None:
            use_socktype = socket.SOCK_DGRAM
        self.socket = socket.socket(socket.AF_UNIX, use_socktype)
        try:
            self.socket.connect(address)
            # it worked, so set self.socktype to the used type
            self.socktype = use_socktype
        except OSError:
            self.socket.close()
            if self.socktype is not None:
                # user didn't specify falling back, so fail
                raise
            use_socktype = socket.SOCK_STREAM
            self.socket = socket.socket(socket.AF_UNIX, use_socktype)
            try:
                self.socket.connect(address)
                # it worked, so set self.socktype to the used type
                self.socktype = use_socktype
            except OSError:
                self.socket.close()
                raise

    def createSocket(self):
        address = self.address
        socktype = self.socktype

        if isinstance(address, str):
            self.unixsocket = True
            # Syslog server may be unavailable during handler initialisation.
            # C's openlog() function also ignores connection errors.
            # Moreover, we ignore these errors while logging, so it not worse
            # to ignore it also here.
            try:
                self._connect_unixsocket(address)
            except OSError:
                pass
        else:
            self.unixsocket = False
            if socktype is None:
                socktype = socket.SOCK_DGRAM
            host, port = self.address
            ress = socket.getaddrinfo(host, port, 0, socktype)
            if not ress:
                raise OSError("getaddrinfo returns an empty list")

            for res in ress:
                af, socktype, proto, _, sa = res
                err = sock = None
                try:
                    sock = socket.socket(af, socktype, proto)
                    if socktype == socket.SOCK_STREAM:
                        sock.connect(sa)
                    break
                except OSError as exc:
                    err = exc
                    if sock is not None:
                        sock.close()
            if err is not None:
                raise err
            self.socket = sock
            self.socktype = socktype

    def encodePriority(self, facility, priority):
        """
        Encode the facility and priority. You can pass in strings or
        integers - if strings are passed, the facility_names and
        priority_names mapping dictionaries are used to convert them to
        integers.
        """
        if isinstance(facility, str):
            facility = self.facility_names[facility]
        if isinstance(priority, str):
            priority = self.priority_names[priority]
        return (facility << 3) | priority

    def close(self):
        """
        Closes the socket.
        """
        self.acquire()
        try:
            sock = self.socket
            if sock:
                self.socket = None
                sock.close()
            logging.Handler.close(self)
        finally:
            self.release()

    def mapPriority(self, levelName):
        """
        Map a logging level name to a key in the priority_names map.
        This is useful in two scenarios: when custom levels are being
        used, and in the case where you can't do a straightforward
        mapping by lowercasing the logging level name because of locale-
        specific issues (see SF #1524081).
        """
        return self.priority_map.get(levelName, "warning")

    ident = ''          # prepended to all messages
    append_nul = True   # some old syslog daemons expect a NUL terminator

    def emit(self, record):
        """
        Emit a record.

        The record is formatted, and then sent to the syslog server. If
        exception information is present, it is NOT sent to the server.
        """
        try:
            msg = self.format(record)
            if self.ident:
                msg = self.ident + msg
            if self.append_nul:
                msg += '\000'

            # We need to convert record level to lowercase, maybe this will
            # change in the future.
            prio = '<%d>' % self.encodePriority(self.facility,
                                                self.mapPriority(record.levelname))
            prio = prio.encode('utf-8')
            # Message is a string. Convert to bytes as required by RFC 5424
            msg = msg.encode('utf-8')
            msg = prio + msg

            if not self.socket:
                self.createSocket()

            if self.unixsocket:
                try:
                    self.socket.send(msg)
                except OSError:
                    self.socket.close()
                    self._connect_unixsocket(self.address)
                    self.socket.send(msg)
            elif self.socktype == socket.SOCK_DGRAM:
                self.socket.sendto(msg, self.address)
            else:
                self.socket.sendall(msg)
        except Exception:
            self.handleError(record)



from django.db.models import Field, Func, Lookup


class SimilarityLookup(Lookup):
    lookup_name = 'similar'

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        res = '{} %%> {}'.format(lhs, rhs), params
        return res


Field.register_lookup(SimilarityLookup)


class WordSimilarity(Func):
    template = '%(expressions)s'
    arg_joiner = '<<->'



import json
import os
from itertools import chain
from pprint import pformat
from urllib.parse import unquote, urlencode

from docutils.parsers.rst import Directive
from jinja2 import Template

from .parse_mixin import ParseMixin

TMPL = Template(
    '''
.. sourcecode:: http

   {{method|upper}} {{path}}{% if query %}?{{query}}{% endif %} HTTP/1.1{% if payload is not none %}
   Content-Type: text/javascript

   {{payload | indent(3, False)}}{% endif %}


.. sourcecode:: http

   HTTP/1.1 {{status_code}} {{status_text}}
   Content-Type: text/javascript

   {{response | indent(3, False)}}
'''
)


class HttpLog(Directive, ParseMixin):
    required_arguments = 1
    has_content = True

    def format(self, s):
        if isinstance(s, str):
            return s
        elif s is not None:
            return pformat(s, indent=1)

    def add_record(self, record):
        record['response'] = self.format(record['response'])
        record['payload'] = self.format(record['payload'])
        query = record.get('query') or ''
        if isinstance(query, str):
            record['query'] = unquote(query)
        elif isinstance(query, dict):
            record['query'] = unquote(urlencode(query))

        return self.parse_lines(TMPL.render(**record).split('\n'))

    def run(self):
        artifact_path = os.environ.get('DOCS_ROOT', './.doc')
        full_path = '{}.json'.format(os.path.join('../', artifact_path, self.arguments[0]))
        with open(full_path) as f:
            records = json.load(f)
        ret = chain(*(self.add_record(record) for record in records))
        return list(ret)


def setup(app):
    app.add_directive('http_log', HttpLog)
    return {'version': '0.0.1'}



"""
Returns monitoring certificate expiration
"""
import asyncio
import logging
from datetime import datetime

from fan_tools.unix import asucc

log = logging.getLogger('monitoring.certs')

DAY_LOOP = 24 * 60 * 60
FMT = '%b %d %H:%M:%S %Y %Z'
CERT_CMD = (
    'echo | openssl s_client -connect {}:443 -servername {} -showcerts 2>/dev/null'  # noqa
    ' | openssl x509 -inform pem -noout -enddate'
)


CERT_HOSTS = ['gettipsi.com', 'proofnetwork.io']


async def get_certs_metrics(cert_hosts):
    metrics = {}
    for host in cert_hosts:
        cmd = CERT_CMD.format(host, host)
        log.debug(f'Run cmd: {cmd}')
        _code, stdout, stderr = await asucc(cmd)
        resp = stdout[0].split('=')[1]
        # notAfter=Dec 25 23:59:59 2018 GMT
        d = datetime.strptime(resp, FMT)
        days_remain = (d - datetime.now()).days
        log.debug(f'Certs response {host} => {days_remain} {resp}')
        metric_name = f'certificate{{config="{host}"}}'
        metrics[metric_name] = days_remain
    return metrics


async def update_certs_loop(update_metrics, hosts):
    while True:
        try:
            metrics = await get_certs_metrics(hosts)
            local_names = update_metrics(metrics)
            log.debug(f'CERTS: {metrics} => {local_names}')
            await asyncio.sleep(DAY_LOOP)
        except Exception:
            log.exception('While get certificates')
            await asyncio.sleep(10)



import argparse
import os
import re
import sys

import yaml

parser = argparse.ArgumentParser(description='fill yaml with environment variables')
parser.add_argument('src', help='source yaml file')
parser.add_argument('dst', help='destination yaml file')


R = re.compile('%\{(.*?)\}')
P = '%{{{}}}'.format
D = os.path.dirname(__file__)


def env(s):
    for k in R.findall(s):
        try:
            s = s.replace(P(k), os.environ.get(k))
        except TypeError as e:
            print('Cannot find env variable: {}'.format(k), file=sys.stderr)
            raise e
    return s


def set_fields(dct):
    for k, v in dct.items():
        if not v and k in os.environ:
            dct[k] = os.environ[k]


def traverse(obj):
    t = type(obj)
    if t == dict:
        return {traverse(k): traverse(v) for k, v in obj.items()}
    elif t == list:
        return list(map(traverse, obj))
    elif t == str:
        return env(obj)
    return obj


def main():
    args = parser.parse_args()
    with open(args.src) as src:
        y = yaml.load(src)
    y = traverse(y)
    with open(args.dst, 'w') as dst:
        yaml.dump(y, dst)


if __name__ == '__main__':
    main()



from django.db.models import CharField, IntegerField, Lookup, Transform
from django.utils.translation import gettext_lazy as _
from django.core.validators import _lazy_re_compile, RegexValidator


class LTreeLabelField(CharField):
    description = _("PostgreSQL ltree label (up to %(max_length)s)")

    default_validators = [
        RegexValidator(
            _lazy_re_compile(r'^[a-zA-Z0-9_]+\Z'),
            _("Enter a valid 'ltree label' consisting of letters, numbers or underscores."),
            'invalid'
        )
    ]

    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = 255
        kwargs['editable'] = False
        kwargs['unique'] = True
        self.allow_unicode = False

        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()

        del kwargs['max_length']
        del kwargs['editable']
        del kwargs['unique']

        return name, path, args, kwargs


class LTreeLabelPathField(CharField):
    description = _("PostgreSQL ltree label path (up to %(max_length)s)")

    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = 2047
        kwargs['editable'] = False
        super().__init__(*args, **kwargs)

    def db_type(self, connection):
        return 'ltree'

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs['max_length']
        del kwargs['editable']
        return name, path, args, kwargs


class LTreeDescendants(Lookup):
    lookup_name = 'descendants'

    def as_sql(self, qn, connection):
        lhs, lhs_params = self.process_lhs(qn, connection)
        rhs, rhs_params = self.process_rhs(qn, connection)
        params = lhs_params + rhs_params
        return '%s <@ %s' % (lhs, rhs), params


class LTreeNlevel(Transform):
    lookup_name = 'nlevel'

    def as_sql(self, compiler, connection):
        lhs, params = compiler.compile(self.lhs)
        return 'nlevel(%s)' % lhs, params

    @property
    def output_field(self):
        return IntegerField()


LTreeLabelPathField.register_lookup(LTreeDescendants)
LTreeLabelPathField.register_lookup(LTreeNlevel)



from rest_framework.renderers import JSONRenderer


class ApiRenderer(JSONRenderer):

    def render(
        self,
        data,
        media_type=None,
        renderer_context=None,
    ):
        wrapper = {
            'version': '001',
        }
        # move error to the root level
        if hasattr(data, 'get') and data.get('error'):
            wrapper['error'] = data['error']
            # UserWrapper support
            wrapper['status_code'] = data['error']['status']
            wrapper['error_message'] = data['error']['message']

            del data['error']

        if data is not None:
            wrapper['data'] = data

        return super().render(
            wrapper,
            media_type,
            renderer_context,
        )



#!/usr/bin/python3
import argparse
import logging
import os
import time

from fan_tools.unix import run

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s [%(levelname)s] %(name)s: %(message)s')
log = logging.getLogger('run_filebeat')

parser = argparse.ArgumentParser(description='run filebeat')
parser.add_argument(
    '-e',
    '--env',
    dest='env',
    action='append',
    default=[],
    help='env vars to check: -e HOST_TYPE=amazon',
)
parser.add_argument(nargs='?', dest='yaml_template', default='./ext_configs/filebeat/config.yaml')
args = parser.parse_args()


def environment_ok():
    required_env = dict([x.split('=') for x in args.env])
    for key, value in required_env.items():
        log.info('Check: {} {}'.format(key, value))
        env_value = os.environ.get(key, None)
        if env_value != value:
            log.info('Environment is not ok: {} => {} != {} '.format(key, value, env_value))
            return False
    return True


def main():
    if not environment_ok():
        while 1:
            time.sleep(9 ** 10)
    if not os.path.exists(args.yaml_template):
        log.exception('Cannot find file: {}'.format(args.yaml_template))
        exit(1)
    run('fan_env_yaml {} /tmp/filebeat.yml'.format(args.yaml_template))
    log.info('Start filebeat')
    out = run('/usr/bin/filebeat -e -c /tmp/filebeat.yml')  # NOQA
    log.exception('Filebeat exited: {}'.format(out))


if __name__ == '__main__':
    main()



from opentelemetry import trace
from opentelemetry.exporter.jaeger.thrift import JaegerExporter
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import SimpleSpanProcessor

from fan_tools.otel import instrument_django, instrument_logging, instrument_psycopg2


def setup_jaeger_tracer(env, service, host, port):
    resource = Resource(attributes={'service.name': service, 'env': env, 'dd.service': service})
    trace.set_tracer_provider(TracerProvider(resource=resource))
    trace.get_tracer_provider().get_tracer(__name__)

    jaeger_exporter = JaegerExporter(
        agent_host_name=host,
        agent_port=port,
    )

    trace.get_tracer_provider().add_span_processor(SimpleSpanProcessor(jaeger_exporter))


INS_DJANGO = [instrument_django, instrument_logging, instrument_psycopg2]


def enable_otel(env, service, host='otlp', port=6831, instrumentations=INS_DJANGO):
    for instrument in instrumentations:
        instrument()
    setup_jaeger_tracer(env, service, host, port)



import asyncio
import logging
import os
import re
import socket
import subprocess
import time
from asyncio.subprocess import PIPE
from collections import ChainMap
from contextlib import closing, contextmanager, suppress
from pathlib import Path
from typing import Callable, List, Optional, TypeVar, Union


log = logging.getLogger('fan_tools.unix')

Cmd = Union[str, List[str]]
Out = List[str]


class ExecError(Exception):
    def __init__(self, exit_code: int, cmd: Cmd, stdout: Out, stderr: Out):
        self.exit_code = exit_code
        self.cmd = cmd
        self.stdout = stdout
        self.stderr = stderr
        super().__init__()

    def __str__(self) -> str:
        return f'Return: {self.exit_code} Command: {self.cmd}\nStderr: {self.stderr}'

    def __repr__(self) -> str:
        desc = str(self).replace('\n', '\t')
        return f'<ExecException `{desc}`>'


def _prepare(out: str) -> Out:
    out: Out = out.decode('utf8').strip('\n').split('\n')
    if out == ['']:
        return []
    return out


def run(command):
    '''
    Run command in shell, accepts command construction from list
    Return (return_code, stdout, stderr)
    stdout and stderr - as list of strings
    '''
    if isinstance(command, list):
        command = ' '.join(command)
    out = subprocess.run(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return (out.returncode, _prepare(out.stdout), _prepare(out.stderr))


def succ(cmd, check_stderr=False, stdout=None, stderr=None):
    '''
    Alias to run with check return code and stderr
    '''
    code, out, err = run(cmd)

    # Because we're raising error, sometimes we want to process stdout/stderr after catching error
    # so we're copying these outputs if required
    if stdout is not None:
        stdout[:] = out
    if stderr is not None:
        stderr[:] = err

    if code != 0:
        for l in out:
            print(l)
    if code != 0:
        raise ExecError(code, cmd, out, err)
    if check_stderr and err != []:
        raise ExecError(code, cmd, out, err)
    return code, out, err


async def process_pipe(out, pipe, proc, log_fun):
    while True:
        line = await pipe.readline()
        line = line.decode('utf8').strip('\n')
        if line:
            out.append(line)
            log_fun(line)
        if not line and proc.returncode is not None:
            return
        await asyncio.sleep(0)


async def asucc(
    cmd, check_stderr=False, pid_future=None, with_log=True, stdout=None, stderr=None, loop=None
):
    proc = await asyncio.create_subprocess_shell(
        cmd, stdin=PIPE, stderr=PIPE, stdout=PIPE, loop=loop
    )
    if pid_future and not pid_future.done():
        pid_future.set_result(proc.pid)

    # see succ comments for these values
    out = stdout if stdout is not None else []
    err = stderr if stderr is not None else []

    if with_log:
        log_warning = log.warning
        log_debug = log.debug
    else:
        skip = lambda x: x
        log_warning, log_debug = skip, skip

    asyncio.ensure_future(process_pipe(err, proc.stderr, proc, log_warning), loop=loop)
    asyncio.ensure_future(process_pipe(out, proc.stdout, proc, log_debug), loop=loop)

    try:
        code = await proc.wait()
        if code != 0:
            raise ExecError(code, cmd, out, err)
        if check_stderr and err != []:
            raise ExecError(code, cmd, out, err)
        return code, out, err
    except asyncio.CancelledError as e:
        if not proc.returncode:
            log.exception('Going to kill process: [{}] {}. Children first'.format(proc.pid, cmd))
            await asucc('pkill -9 -P {} || true'.format(proc.pid))
            with suppress(ProcessLookupError):
                proc.terminate()
            await asyncio.sleep(0.1)
            if not proc.returncode:
                with suppress(ProcessLookupError):
                    proc.kill()
        raise e


def check_socket(host: str, port: int) -> bool:
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.settimeout(10)
        try:
            conn_ex = sock.connect_ex((host, port))
            if conn_ex == 0:
                return True
            return False
        except:
            return False


T = TypeVar('T')


def wait_result(func: Callable, result: Optional[T], timeout) -> bool:
    t = time.time()
    while True:
        res = func()
        if result is None and res is None:
            return True
        elif res == result:
            return True
        time.sleep(1)
        if (time.time() - t) > timeout:
            return False


def wait_socket(host: str, port: int, timeout=120) -> bool:
    '''
    Wait for socket opened on remote side. Return False after timeout
    '''
    return wait_result(lambda: check_socket(host, port), True, timeout)


def wait_no_socket(host: str, port: int, timeout=120) -> bool:
    return wait_result(lambda: check_socket(host, port), False, timeout)


def interpolate_sysenv(line: str, defaults={}) -> str:
    '''
    Format line system environment variables + defaults
    '''
    map = ChainMap(os.environ, defaults)
    return line.format(**map)


def source(fname: str):
    '''
    Acts similar to bash 'source' or '.' commands.
    '''
    rex = re.compile('(?:export |declare -x )?(.*?)="(.*?)"')
    out = succ('source {} && export'.format(fname))[1]
    out = [x for x in out if 'export' in x or 'declare' in x]
    out = {k: v for k, v in [rex.match(x).groups() for x in out if rex.match(x)]}
    for k, v in out.items():
        os.environ[k] = v


@contextmanager
def cd(dir_name: Union[str, Path]):
    """
    do something in other directory and return back after block ended
    """
    old_path = os.path.abspath('.')
    os.chdir(dir_name)
    try:
        yield
        os.chdir(old_path)
    except Exception:
        os.chdir(old_path)
        raise



from setuptools import find_packages, setup

with open('fan_tools/__init__.py', 'r') as f:
    for line in f:
        if line.startswith('__version__'):
            version = line.strip().split('=')[1].strip(' \'"')
            break

with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()

SANIC = 'sanic'

setup(
    name='fan_tools',
    packages=find_packages(exclude=('tests', 'django_tests.*', 'django_tests')),
    version=version,
    description='Various python stuff: testing, aio helpers, etc',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='cybergrind',
    author_email='cybergrind@gmail.com',
    url='http://github.com/micro-fan/fan_tools',
    keywords=['testing', 'asyncio'],
    install_requires=['pyyaml>=3.12', 'python-json-logger>=0.1.5'],
    extras_require={
        'doc_utils': [
            'rest_framework_dyn_serializer>=1.3.*',
            'docutils',
            'djangorestframework==3.12.*',
        ],
        'logging': ['safe-logger>=1.2.1', 'python-json-logger==0.1.7'],
        'aio_utils': ['aiopg', 'psycopg2-binary'],
        'monitoring': [SANIC, 'starlette'],
        's3_backup': ['boto3'],
        'gcloud_backup': ['google-cloud-storage'],
        'gitlab_monitoring': ['python-gitlab==1.0.2', SANIC],
        'image_utils': ['Pillow'],
        'otel': ['opentelemetry-instrumentation-django', 'opentelemetry-instrumentation-psycopg2'],
        'jaeger': ['opentelemetry-exporter-jaeger-thrift'],
    },
    tests_require=['pytest==3.1.3'],
    entry_points={
        'console_scripts': [
            'fan_env_yaml=fan_tools.scripts.fan_env_yaml:main',
            'fan_ci_script=fan_tools.scripts.fan_ci_script:main',
            'fan_wait=fan_tools.scripts.fan_tools_wait:main',
            'run_filebeat=fan_tools.scripts.run_filebeat:main',
            'doc_serializer=fan_tools.doc_utils.fan_sphinx.dyn:main',
            'fan_s3_backup=fan_tools.backup.s3:main',
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
    ],
)



import asyncio
import logging

from psycopg2.extras import DictCursor

import aiopg


class DbRecordsProcessorWorker:
    """
    Asyncio worker which wait for new records in pg db table and process them.
    """

    MAX_LOOP = 12

    def __init__(self, dsn, worker_class, max_loop=MAX_LOOP):
        self.name = worker_class.__name__
        self.log = logging.getLogger(f'processor.{self.name}')

        self.dsn = dsn
        self.worker = worker_class(self.log)
        self.max_loop = max_loop

        self.debug('init processor')
        self.loop = asyncio.get_event_loop()
        self.closed = self.loop.create_future()

    def debug(self, msg):
        self.log.debug(f'[{self.name}] {msg}')

    def info(self, msg):
        self.log.info(f'[{self.name}] {msg}')

    async def on_start(self):
        self.debug(f'Starting... {self.dsn}')
        self.pool = await aiopg.create_pool(self.dsn)
        self.loop_timer = self.loop.call_soon(self.main_loop)
        self.debug('Started')

    def handle_errors(self, fut):
        if fut.exception():
            self.closed.set_exception(fut.exception())

    def main_loop(self):
        self.debug('Main loop')
        self.loop.create_task(self._inner()).add_done_callback(self.handle_errors)

    async def _inner(self):
        async with self.pool.acquire() as conn:
            await self.worker.process_records(conn)
            try:
                await asyncio.wait_for(self.notify_wait(conn), self.max_loop)
            except asyncio.TimeoutError:
                pass
        self.loop.call_soon(self.main_loop)

    async def notify_wait(self, conn):
        async with conn.cursor() as cursor:
            try:
                await cursor.execute(f'LISTEN {self.worker.pg_channel}')
                await asyncio.wait_for(conn.notifies.get(), self.max_loop)
            except asyncio.TimeoutError:
                pass

    def run(self):
        def check_error(fut):
            e = fut.exception()
            if e:
                logging.exception('Terminate: {}'.format(e))
                loop = asyncio.get_event_loop()
                loop.stop()
                loop.close()

        self.info(f'Start {self.name}')
        loop = asyncio.get_event_loop()
        task = loop.create_task(self.on_start())
        task.add_done_callback(check_error)
        self.debug('run loop')
        loop.run_until_complete(self.closed)


async def dict_query(conn, query, params):
    async with conn.cursor(cursor_factory=DictCursor) as cur:
        await cur.execute(query, params)
        return [dict(x) for x in await cur.fetchall()]


async def sql_update(conn, query, params):
    async with conn.cursor(cursor_factory=DictCursor) as cur:
        await cur.execute(query, params)



import os
import asyncio
import datetime
import logging

from fan_tools.unix import asucc


class BackupMonitoring:
    log = logging.getLogger('BackupMonitoring')
    LOOP = 24 * 60 * 60
    RETRY_TIMEOUT = 5 * 60
    NAME_PATTERN = '%Y%m%d_%H%M.gz'

    def __init__(self, prefix, metric_name, backup_command, enabled=True):
        self.prefix = prefix
        self.metric_name = metric_name
        self.backup_command = backup_command
        self.enabled = enabled
        self.last = datetime.datetime.min

    async def run_monitoring(self, update_metrics):
        self.update_metrics = update_metrics
        while True:
            try:
                self.set_latest_timestamp()
                break
            except Exception:
                self.log.exception('Cannot get latest backup. Sleep 5 seconds')
                await asyncio.sleep(5)

        while True:
            try:
                wait_seconds = (self.last - datetime.datetime.utcnow()).total_seconds() + self.LOOP
                if wait_seconds <= 0:
                    await self.perform_backup()
                    self.set_latest_timestamp()
                    self.log.debug(f'Sleep after backup for: {self.LOOP}')
                    await asyncio.sleep(self.LOOP)
                else:
                    self.log.debug(f'Sleep without backup for: {wait_seconds}. Last: {self.last}')
                    await asyncio.sleep(wait_seconds)
            except Exception:
                self.log.exception(f'During perform_backup call. Sleep for {self.RETRY_TIMEOUT}')
                await asyncio.sleep(self.RETRY_TIMEOUT)

    def set_latest_timestamp(self):
        flist = self.get_backups_list()
        if len(flist) == 0:
            return
        dates = sorted([datetime.datetime.strptime(f, self.NAME_PATTERN) for f in flist])
        self.last = dates[-1]
        self.call_update()

    def call_update(self):
        if self.last > datetime.datetime.min:
            self.update_metrics({self.metric_name: self.last.timestamp()})

    def get_backups_list(self):
        raise NotImplementedError

    async def perform_upload(self, src_name, dst_name):
        raise NotImplementedError

    async def perform_backup(self, skip_upload=False):
        if not self.enabled:
            self.log.debug('Backups disabled. Skip action. To enable set ENABLE_BACKUP=true')
            return

        code, out, err = await asucc(self.backup_command)
        self.log.debug(f'Response: {code} => {out} Err: {err}')
        assert code == 0, code
        fname = out[0].strip()
        src_name = f'/dumps/{fname}'
        dst_name = os.path.join(self.prefix, fname)

        if skip_upload:
            return

        self.perform_upload(src_name, dst_name)



#!/usr/bin/env python3
"""
Export items in prometheus format.
"""
import logging
import os
from collections import ChainMap
from functools import partial

from sanic import Sanic, response

log = logging.getLogger('fan_monitoring')


class MetricsServer:
    """
    You you should add monitoring functions with add_task
    Your update function will receive update_metrics function with all additional args
    """
    def __init__(self, app, status_metric='running{example_var="default_env"}'):
        self.app = app
        self.app.add_route(self.expose_metrics, '/metrics')
        self.metrics = f'{status_metric} 1'
        self.all_names = [{status_metric: 1}]

    def update_metrics(self, names, metrics):
        for name in names:
            if name not in metrics:
                names[name] = 0

        for name, value in metrics.items():
            names[name] = value

        mlist = [f'{name} {value}' for name, value in ChainMap(*self.all_names).items()]
        self.metrics = '\n'.join(mlist)
        return names

    def gen_update_func(self):
        local_names = {}
        self.all_names.append(local_names)
        return partial(self.update_metrics, local_names)

    def add_task(self, loop_func, *args, **kwargs):
        self.app.add_task(loop_func(self.gen_update_func(), *args, **kwargs))

    async def expose_metrics(self, request):
        return response.text(self.metrics)


def main():
    from fan_tools.mon_server.certs import update_certs_loop
    app = Sanic()
    mserver = MetricsServer(app)
    mserver.add_task(update_certs_loop, hosts=['gettipsi.com', 'proofnetwork.io'])
    app.run(host='0.0.0.0', port=os.environ.get('MONITORING_PORT'))


if __name__ == '__main__':
    main()
